// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:wyatt_crud_bloc/src/core/enums/where_query_type.dart';
import 'package:wyatt_crud_bloc/src/core/mixins/operation.dart';
import 'package:wyatt_crud_bloc/src/domain/entities/object_model.dart';
import 'package:wyatt_crud_bloc/src/domain/entities/query.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/create.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/delete.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/delete_all.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/get.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/get_all.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/params/update_parameters.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/query.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/update.dart';
import 'package:wyatt_crud_bloc/src/domain/usecases/update_all.dart';
import 'package:wyatt_crud_bloc/src/features/crud/blocs/crud_base_cubit/crud_base_cubit.dart';

/// {@template crud_cubit}
/// Cubit that handles CRUD operations.
/// {@endtemplate}
abstract class CrudCubit<Model extends ObjectModel> extends CrudBaseCubit {
  /// {@macro crud_cubit}
  CrudCubit() : super();

  /// Create operation.
  /// Can be create.
  CreateOperation<Model, dynamic>? get createOperation;

  /// Read operation.
  /// Can be get, getAll, query.
  ReadOperation<Model, dynamic, dynamic>? get readOperation;

  /// Update operation.
  /// Can be update, updateAll.
  UpdateOperation<Model, dynamic>? get updateOperation;

  /// Delete operation.
  /// Can be delete or deleteAll.
  DeleteOperation<Model, dynamic>? get deleteOperation;

  Expected? _checkOperation<Expected>(dynamic operation) {
    if (operation == null) {
      return null;
    }
    if (operation is! Expected) {
      return null;
    }
    return operation;
  }

  FutureOr<void> create(Model model) async {
    if (_checkOperation<Create<Model>>(createOperation) != null) {
      return _create(model);
    }
  }

  FutureOr<void> read({String? id, List<QueryInterface>? conditions}) async {
    if (_checkOperation<Get<Model>>(readOperation) != null && id != null) {
      return _get(id);
    }
    if (_checkOperation<GetAll<Model>>(readOperation) != null) {
      return _getAll();
    }
    if (_checkOperation<Query<Model>>(readOperation) != null &&
        conditions != null) {
      return _query(conditions);
    }
    if (_checkOperation<Query<Model>>(readOperation) != null &&
        conditions == null) {
      return _getAll();
    }
  }

  FutureOr<void> update(
    UpdateParameters<Model>? single,
    Map<String, dynamic>? all,
  ) async {
    if (_checkOperation<Update<Model>>(updateOperation) != null &&
        single != null) {
      return _update(single);
    }

    if (_checkOperation<UpdateAll<Model>>(updateOperation) != null &&
        all != null) {
      return _updateAll(all);
    }
  }

  FutureOr<void> delete({String? id}) async {
    if (_checkOperation<Delete<Model>>(deleteOperation) != null && id != null) {
      return _delete(id);
    }
    if (_checkOperation<DeleteAll<Model>>(deleteOperation) != null) {
      return _deleteAll();
    }
  }

  FutureOr<void> _create(Model model) async {
    final crud = _checkOperation<Create<Model>>(createOperation);
    if (crud == null) {
      return;
    }

    final stateCopy = state;
    emit(const CrudLoading());
    final result = await crud.call(model);
    emit(
      result.fold(
        (_) {
          if (stateCopy is CrudLoaded<Model?>) {
            if (stateCopy.data == null) {
              return CrudLoaded<Model?>(model);
            }
            if (stateCopy.data!.id == model.id) {
              return CrudLoaded<Model?>(model);
            }

            return stateCopy;
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            if (stateCopy.data.isEmpty) {
              return CrudListLoaded<Model?>([model]);
            }
            final List<Model?> lst = stateCopy.data.toList()..add(model);

            return CrudListLoaded<Model?>(lst);
          }

          return const CrudOkReturn();
        },
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> _delete(String id) async {
    final crud = _checkOperation<Delete<Model>>(deleteOperation);
    if (crud == null) {
      return;
    }

    final stateCopy = state;
    emit(const CrudLoading());
    final result = await crud.call(id);
    emit(
      result.fold(
        (_) {
          if (stateCopy is CrudLoaded<Model?>) {
            if (stateCopy.data?.id == id) {
              return CrudLoaded<Model?>(null);
            }

            return stateCopy;
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            return CrudListLoaded<Model?>(
              stateCopy.data.where((element) => element?.id != id).toList(),
            );
          }

          return const CrudOkReturn();
        },
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> _deleteAll() async {
    final crud = _checkOperation<DeleteAll<Model>>(deleteOperation);
    if (crud == null) {
      return;
    }
    final stateCopy = state;
    emit(const CrudLoading());
    final result = await crud.call(null);
    emit(
      result.fold(
        (_) {
          if (stateCopy is CrudLoaded<Model?>) {
            return CrudLoaded<Model?>(null);
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            return CrudListLoaded<Model?>(const []);
          }

          return const CrudOkReturn();
        },
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> _get(String id) async {
    final crud = _checkOperation<Get<Model>>(readOperation);
    if (crud == null) {
      return;
    }
    emit(const CrudLoading());
    final result = await crud.call(id);
    emit(
      result.fold(
        CrudLoaded<Model?>.new,
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> _getAll() async {
    final crud = _checkOperation<GetAll<Model>>(readOperation);
    if (crud == null) {
      return;
    }
    emit(const CrudLoading());
    final result = await crud.call(null);
    emit(
      result.fold(
        CrudListLoaded<Model?>.new,
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> _query(List<QueryInterface> conditions) async {
    final crud = _checkOperation<Query<Model>>(readOperation);
    if (crud == null) {
      return;
    }

    emit(const CrudLoading());
    final result = await crud.call(conditions);
    emit(
      result.fold(
        CrudListLoaded<Model?>.new,
        (error) => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> _update(UpdateParameters<Model> param) async {
    final crud = _checkOperation<Update<Model>>(updateOperation);
    if (crud == null) {
      return;
    }

    final stateCopy = state;
    emit(const CrudLoading());
    final result = await crud.call(param);
    emit(
      await result.foldAsync(
        (_) async {
          if (stateCopy is CrudLoaded<Model?>) {
            if (stateCopy.data?.id == param.id) {
              // Same object, need to update actual stateCopy
              final crudGet = _checkOperation<Get<Model>>(readOperation);
              if (crudGet == null) {
                // No read operation, can't update stateCopy.
                return stateCopy;
              }
              final newVersion = await crudGet.call(param.id);
              if (newVersion.isOk) {
                return CrudLoaded<Model?>(newVersion.ok);
              }
            }
            return stateCopy;
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            final bool listContains =
                stateCopy.data.any((element) => element?.id == param.id);
            if (listContains) {
              // Loaded objects contains the modified object.

              final crudGet = _checkOperation<Get<Model>>(readOperation);
              if (crudGet == null) {
                // No read operation, can't update stateCopy.
                return stateCopy;
              }
              final newVersion = await crudGet.call(param.id);
              if (newVersion.isOk) {
                final newList = stateCopy.data
                    .where(
                      (element) => element?.id != param.id,
                    )
                    .toList();
                return CrudListLoaded<Model?>(newList + [newVersion.ok]);
              }
            }
            return stateCopy;
          }
          return const CrudOkReturn();
        },
        (error) async => CrudError(error.toString()),
      ),
    );
  }

  FutureOr<void> _updateAll(Map<String, Object?> param) async {
    final crud = _checkOperation<UpdateAll<Model>>(updateOperation);
    if (crud == null) {
      return;
    }

    final stateCopy = state;
    emit(const CrudLoading());
    final result = await crud.call(param);
    emit(
      await result.foldAsync(
        (_) async {
          if (stateCopy is CrudLoaded<Model?>) {
            // Same object, need to update actual stateCopy
            final crudGet = _checkOperation<Get<Model>>(readOperation);
            if (crudGet == null) {
              // No read operation, can't update stateCopy.
              return stateCopy;
            }
            final actualId = stateCopy.data?.id;
            final newVersion = await crudGet.call(actualId ?? '');
            if (newVersion.isOk) {
              return CrudLoaded<Model?>(newVersion.ok);
            }
            return stateCopy;
          }
          if (stateCopy is CrudListLoaded<Model?>) {
            final crudQuery = _checkOperation<Query<Model>>(readOperation);
            if (crudQuery == null) {
              // No read operation, can't update stateCopy.
              return stateCopy;
            }
            // Load all id to retrieve exactly same object
            // (not all because previous stateCopy can be a query result)
            final List<String?> ids = stateCopy.data
                .map(
                  (e) => e?.id,
                )
                .toList();
            final result = await crudQuery.call([
              WhereQuery(
                WhereQueryType.whereIn,
                'id',
                ids,
              ),
            ]);
            if (result.isOk) {
              return CrudListLoaded<Model?>(result.ok ?? []);
            }
            return stateCopy;
          }
          return const CrudOkReturn();
        },
        (error) async => CrudError(error.toString()),
      ),
    );
  }
}
